local JD_RADIUS = 10
local JD_STEP = (JD_RADIUS * 2) + 1

local info = function(value)
  digiline_send('engine_control', value)
end

local jd_error = function(errormsg)
  local result = 'unknown'
  if type(errormsg) ~= 'string' or #errormsg < mem.jd_error_match_min_length then
    -- Shortcut if errormsg is not string or shorter than any of matches
    return result
  end
  for k,v in pairs(mem.jd_error_match) do
    if string.sub(event.msg.msg, 1, #v) == v then
      result = k
      break;
    end
  end
  return result
end

local jd_command = function(cmd, arg)
  if cmd ~= 'reset' and cmd ~= 'set' then
    -- reset and set wont send replies, every other commad does
    mem.processing = true
    info('Processing jumpdrive command ' .. cmd)
  end
  if cmd == 'set' then
    digiline_send('jumpdrive', { command = "set", key = 'x', value = arg.x })
    digiline_send('jumpdrive', { command = "set", key = 'y', value = arg.y })
    digiline_send('jumpdrive', { command = "set", key = 'z', value = arg.z })
    if arg.r ~= nil then
      digiline_send('jumpdrive', { command = "set", key = 'radius', value = arg.r })
    end
    mem.jd.initialized = true
  else
    -- assume no args, currently only `command` used for jd functions other than `set`
    digiline_send('jumpdrive', { command = cmd })
  end
end

local reset = function()
  mem.processing = false
  if mem.offset then
    -- Coords to previous location
    mem.offset = { x = -mem.offset.x, y = -mem.offset.y, z = -mem.offset.z }
  else
    -- Coords to origin
    mem.offset = { x = 0, y = 0, z = 0 }
  end
  mem.jd = {
    verified = false,
    initialized = false,
  }
  jd_command('get')
end

local offsetpos = function(pos, offset)
  return {
    x = pos.x + offset.x,
    y = pos.y + offset.y,
    z = pos.z + offset.z,
    radius = (pos.radius and offset.radius) and (pos.radius + offset.radius) or nil,
  }
end

local pos2str = function(pos)
  if not pos.x or not pos.y or not pos.z then
    return '<invalid pos>'
  end
  return pos.x .. ',' .. pos.y .. ',' .. pos.z
end

local jumpdrive = {}

jumpdrive.on_jumpdrive_data = function(power, radius, pos, target, distance, power_req)
  mem.jd.x = pos.x
  mem.jd.y = pos.y
  mem.jd.z = pos.z
  if not mem.jd.initialized then
    jd_command('set', offsetpos(mem.jd, mem.offset))
    info('Initializing jumpdrive, plese wait...')
    interrupt(2)
  else
    info('Jumpdrive at ' .. pos2str(mem.jd)
      .. '.\n - Target: ' .. pos2str(target)
      .. '\n - Radius: ' .. radius
      .. '\n - Power: ' .. power
      .. '\n - Required power: ' .. power_req .. ' for distance ' .. distance
      .. '\n - Excess power: ' .. (power - power_req)
    )
  end
  mem.processing = false
end

jumpdrive.on_simulation_complete = function(success, msg)
  local err = jd_error(msg)
  mem.processing = false
  if success then
    info('Jumpdrive simulation success, ready for jump. Jumpdrive reply: ' .. (msg and msg or 'OK'))
    mem.jd.verified = true
  else
    info('Jumpdrive simulation failed, cannot jump. Error was: ' .. err)
  end
end

jumpdrive.on_jump_complete = function(success, time)
  reset()
  if success then
    info('Jumpdrive jump success, ready for jump. Jumpdrive reply: ' .. (msg and msg or 'OK') .. '\n - Hit green check again to execute jump')
  else
    info('Jumpdrive jump failed, cannot jump. Error was: ' .. err)
  end
end

local on_digiline = {
  engine_control = function(msg)

    if mem.processing then
      info('Confirmation for last command not yet received... please wait')
      return
    end

    if msg == 'up' then
      mem.jd.verified = false
      mem.offset.z = mem.offset.z + JD_STEP
      info('Reloading 4 seconds after last action...')
      interrupt(4)
    elseif msg == 'down' then
      mem.jd.verified = false
      mem.offset.z = mem.offset.z - JD_STEP
      info('Reloading 4 seconds after last action...')
      interrupt(4)
    elseif msg == 'left' then
      mem.jd.verified = false
      mem.offset.x = mem.offset.x - JD_STEP
      info('Reloading 4 seconds after last action...')
      interrupt(4)
    elseif msg == 'right' then
      mem.jd.verified = false
      mem.offset.x = mem.offset.x + JD_STEP
      info('Reloading 4 seconds after last action...')
      interrupt(4)
    elseif msg == 'enter' then
      jd_command('set', offsetpos(mem.jd, mem.offset))
      if mem.jd.verified then
        jd_command('Xjump')
      else
        jd_command('show')
      end
    elseif msg == 'back' then
      jd_command('reset')
    end

  end,

  jumpdrive = function(msg)
    if msg.target and msg.position then
      jumpdrive.on_jumpdrive_data(
        math.floor(msg.powerstorage),
        msg.radius,
        msg.position,
        msg.target,
        math.ceil(msg.distance),
        math.ceil(msg.power_req)
      )
    elseif msg.msg ~= nil and msg.success ~= nil then
      jumpdrive.on_simulation_complete(msg.success, msg.msg)
    elseif msg.time ~= nil and msg.success ~= nil then
      jumpdrive.on_jump_complete(msg.success, msg.time)
    else
      info('Received unknown or corrupted message from jumpdrive')
    end
  end,
}

local initialize_jd_error_parser = function()
  mem.jd_error_match = {
    mapgen = 'Error: mapgen',
    overlap = 'Error: jump into itself',
    blacklist = "Can't jump node",
    vacuum = 'Warning: Jump-target is in vacuum',
    uncharted = 'Warning: Jump-target is in uncharted',
    protected = 'Jump-target is protected',
    obstructed = 'Jump-target is obstructed',
    power = 'Not enough power',
  }
  -- mem.jd_error_match_min_length = math.huge <= was thinking about this but then this is just small luac, let's use some actual number because idk what else this affects?
  mem.jd_error_match_min_length = 99
  for k,v in pairs(mem.jd_error_match) do
    if mem.jd_error_match_min_length > #v then
      mem.jd_error_match_min_length = #v
    end
  end
end

if event.type == 'digiline' then
  on_digiline[event.channel](event.msg)
  digiline_send('lcd', event.msg)
elseif event.type == 'interrupt' then
  jd_command('set', offsetpos(mem.jd, mem.offset))
  jd_command('get')
elseif event.type == 'program' then
  initialize_jd_error_parser()
  reset()
end